import 'dart:async';
import 'dart:convert';

import 'package:flutter/services.dart';
import 'package:flutter_nfc_kit/flutter_nfc_kit.dart';
import 'package:prepaid_lib_flutter/ApduResponse.dart';

class UnikLibFlutter {
  static const MethodChannel _channel = const MethodChannel('unik_lib_flutter');

  static Future<String> get platformVersion async {
    final String version = await _channel.invokeMethod('getPlatformVersion');
    return version;
  }

  static void initUnikLib(String mid, int env) async {
    await _channel
        .invokeMethod('initUnikLib', <String, dynamic>{'mid': mid, 'env': env});
  }

  static Future<bool> getCardInfo(
      List<String> cardUid,
      List<String> cardNumber,
      List<String> balance,
      List<String> bankName,
      List<String> cardAttr,
      List<String> cardInfo,
      {List<String> rrn,
      bool startPooling,
      List<String> beforeBalance}) async {
    if (beforeBalance == null) beforeBalance = List(1);
    bool res = false;
    String apduClSelectApp = await _channel.invokeMethod('apduIsMyCard');
    String apduClCardInfo = await _channel.invokeMethod('apduCardInfo');
    String apduClBalance = await _channel.invokeMethod('apduBalance');
    String apduClCardAttr = await _channel.invokeMethod('apduCardAttr');

    try {
      if (startPooling) {
        NFCTag tag = await FlutterNfcKit.poll();
        cardUid[0] = tag.id;
      }

      var dataSelect = jsonDecode(apduClSelectApp);
      var dataBalance = jsonDecode(apduClBalance);
      var dataCardAttr = jsonDecode(apduClCardAttr);

      List<ApduResponse> listData = dataSelect
          .map<ApduResponse>((result) => new ApduResponse.fromJson(result))
          .toList();
      List<ApduResponse> listApduBalance = dataBalance
          .map<ApduResponse>((result) => new ApduResponse.fromJson(result))
          .toList();
      List<ApduResponse> listApduCardAttr = dataCardAttr
          .map<ApduResponse>((result) => new ApduResponse.fromJson(result))
          .toList();

      for (ApduResponse s in listData) {
        String selectApp = await FlutterNfcKit.transceive(s.apdu);
        if (selectApp == "9000") {
          bankName[0] = s.bankType;
          print("bank name ${bankName[0]}");
        }
      }

      for (ApduResponse b in listApduBalance) {
        if (b.bankType == bankName[0]) {
          String rApduCardAttr = "";

          if (bankName[0] == "MANDIRI") {
            String apduBalance = await FlutterNfcKit.transceive(b.apdu);

            cardInfo[0] = await FlutterNfcKit.transceive(apduClCardInfo);
            for (ApduResponse apduCardAttr in listApduCardAttr) {
              if (apduCardAttr.bankType == bankName[0])
                rApduCardAttr =
                    await FlutterNfcKit.transceive(apduCardAttr.apdu);
            }

            String result = await _channel.invokeMethod(
                'parseBalance', <String, dynamic>{
              'apduBalance': "$apduBalance,${cardInfo[0]}",
              'bankType': bankName[0]
            });
            print("result $result");
            List<String> items = result.split(",");
            balance[0] = items[0];
            beforeBalance[0] = items[0];
            cardNumber[0] = items[1];
            cardAttr[0] = rApduCardAttr;
          } else {
            // todo for bni
            for (ApduResponse apduCardAttr in listApduCardAttr) {
              if (apduCardAttr.bankType == bankName[0]) {
                rApduCardAttr =
                    await FlutterNfcKit.transceive(apduCardAttr.apdu);
              }
            }

            print("apduBalance --> ${b.apdu}");
            String apduBalance = await FlutterNfcKit.transceive(b.apdu);
            print("apduBalance <-- $apduBalance");

            String purse = apduBalance.substring(0, apduBalance.length - 4);
            String crn = rApduCardAttr.substring(0, rApduCardAttr.length - 4);
            String rRrn = b.apdu.substring(14, 30);
            print("rRrn $rRrn");
            print("purse $purse");
            print("crn $crn");

            if (rrn != null) rrn[0] = rRrn;

            String composeBni = await _channel.invokeMethod('composeBNI',
                <String, dynamic>{'apduPurse': purse + crn + rRrn});

            String result = await _channel.invokeMethod(
                'parseBalance', <String, dynamic>{
              'apduBalance': purse,
              'bankType': bankName[0]
            });
            print("result $result");
            List<String> items = result.split(",");
            balance[0] = items[0];
            beforeBalance[0] = items[0];
            cardNumber[0] = items[1];
            cardInfo[0] = composeBni ?? "0";
          }

          print("cardNumber = ${cardNumber[0]}, balance ${balance[0]}");

          res = true;
        }
      }
    } catch (error) {
      res = false;
      print("error.toString() ${error.toString()}");
    }
    return res;
  }

  static Future<bool> updateBalance(
      List<String> status,
      List<String> cardNumber,
      List<String> balance,
      List<String> bankName,
      List<String> beforeBalance) async {
    bool result = false;
    List<String> cardUid = List(1);
    List<String> cardAttr = List(1);
    List<String> cardInfo = List(1);
    List<String> lRrn = List(1);
    String rApduCardAttr = "";

    try {
      bool isSuccess = await getCardInfo(
          cardUid, cardNumber, balance, bankName, cardAttr, cardInfo,
          rrn: lRrn, startPooling: true, beforeBalance: beforeBalance);

      if (isSuccess) {
        if (await doReversalIfExist(cardNumber[0], "", "")) {
          String resUpdate =
              await _channel.invokeMethod('updateBalance', <String, dynamic>{
            'cardNumber': cardNumber[0] ?? "",
            'cardUid': cardUid[0] ?? "",
            'bankType': bankName[0] ?? "",
            'balance': balance[0] ?? "",
            'cardAttr': cardAttr[0] ?? "0",
            'respCardInfo': cardInfo[0] ?? "0"
          }).onError((error, stackTrace) {
            status[0] = error.toString();
          });

          List<String> items = resUpdate.toString().split(",");

          if (items[0] != "400") {
            // todo confirm update
            if (bankName[0] == "MANDIRI") {
              String apduClCert = await _channel.invokeMethod('apduCardCert');
              String rApdu = await FlutterNfcKit.transceive(items[1]);
              if (items[2] == "NEW") {
                String rApduCert = await FlutterNfcKit.transceive(apduClCert);
                rApdu = rApdu.substring(0, rApdu.length - 4) +
                    rApduCert.substring(0, rApduCert.length - 4);
              }

              result = await confirmUpdate(cardUid[0], bankName[0],
                  cardNumber[0], rApdu, status, "", balance[0], "");

              // get balance again for mandiri
              await getCardInfo(
                  cardUid, cardNumber, balance, bankName, cardAttr, cardInfo,
                  startPooling: false);
              print("result confirm on update $result");
            } else {
              // todo update bni
              String crypto = await _channel.invokeMethod('createCrypto',
                  <String, dynamic>{'crypto': items[1], 'rrn': lRrn[0]});

              String rApdu;
              String statusInj;
              for (int i = 3; i != 0; i--) {
                rApdu = await FlutterNfcKit.transceive(crypto);
                print("rApduDataToCardUpdateBni --> $rApdu");
                statusInj = rApdu.substring(rApdu.length - 4, rApdu.length);
                if (statusInj == "9000") {
                  break;
                }
              }

              // get balance again for bni
              await getCardInfo(
                  cardUid, cardNumber, balance, bankName, cardAttr, cardInfo,
                  startPooling: false);

              if (statusInj == "9000") {
                result = await confirmUpdate(cardUid[0], bankName[0],
                    cardNumber[0], rApdu, status, items[2], balance[0], "");
                print("result confirm on update $result");
              } else {
                result = false;
              }
            }
          } else {
            status[0] = items[1];
            result = true;
          }
        }
      }
    } catch (error) {
      status[0] = "Something went wrong";
    }

    return result;
  }

  static Future<bool> doReversalIfExist(
      String cardNumber, String samData, String mode) async {
    var revResult = await _channel.invokeMethod(
        "checkingReversal", <String, dynamic>{
      'cardNumber': cardNumber,
      'samData': samData,
      'mode': mode
    });

    if (revResult == "") {
      return true;
    } else {
      String apdu = await FlutterNfcKit.transceive(revResult.toString());
      return doReversalIfExist(cardNumber, apdu, "REVERSAL_LOOP");
    }
  }

  static Future<bool> confirmUpdate(
      String cardUid,
      String bankType,
      String cardNumber,
      String samData,
      List<String> status,
      String reffNo,
      String balance,
      String writeStatus) async {
    String respConfirm =
        await _channel.invokeMethod('confirm_update', <String, dynamic>{
      'cardUid': cardUid,
      'bankType': bankType,
      'cardNumber': cardNumber,
      'samData': samData,
      'reff_no': reffNo,
      'balance': balance,
      'writeStatus': writeStatus
    });

    List<String> items = respConfirm.toString().split(",");
    if (items[0] == "" || items[0] == "N/A") {
      print("respConfirm ${items[1]}");
      status[0] = items[1];
      return true;
    } else if (items[0].contains("No Pending Balance") ||
        items[0].contains("Success")) {
      status[0] = items[0];
      return true;
    } else {
      String apdu = await FlutterNfcKit.transceive(items[0]);
      String statusInj = apdu.substring(apdu.length - 4, apdu.length);
      return confirmUpdate(cardUid, bankType, cardNumber, apdu, status, reffNo,
          balance, statusInj == "9000" && items[2] == "NEW" ? "COMPLETED" : "");
    }
  }
}
